
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.28.0',
    'branch_nick': 'master',
    'revision_id': 'b661efb29efe34c871e1627610c3f1cff4b36813',
    'revno': 11482
}
